@@ -1,10 +1,17 @@
 ;;; tree-buffer.el --- functions for tree buffers
 
-;; Copyright (C) 2000, 2001 Jesper Nordenberg
+;; Copyright (C) 2000 - 2003 Jesper Nordenberg,
+;;                           Klaus Berndl,
+;;                           Kevin A. Burton,
+;;                           Free Software Foundation, Inc.
 
 ;; Author: Jesper Nordenberg <mayhem@home.se>
+;;         Klaus Berndl <klaus.berndl@sdm.de>
+;;         Kevin A. Burton <burton@openprivacy.org>
 ;; Maintainer: Klaus Berndl <klaus.berndl@sdm.de>
-;; Keywords: java, class, browser
+;;             Kevin A. Burton <burton@openprivacy.org>
+;; Keywords: browser, code, programming, tools, tree
+;; Created: 2000
 
 ;; This program is free software; you can redistribute it and/or modify it under
 ;; the terms of the GNU General Public License as published by the Free Software
@@ -19,8 +26,10 @@
 ;; GNU Emacs; see the file COPYING.  If not, write to the Free Software
 ;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 
+;; $Id: tree-buffer.el,v 1.8 2003/09/18 10:16:07 berndl Exp $
+
 ;;; Commentary:
-;;
+
 ;; Functions for tree buffers.
 ;;
 ;; This file is part of the ECB package which can be found at:
@@ -31,8 +40,6 @@
 ;; For the ChangeLog of this file see the CVS-repository. For a complete
 ;; history of the ECB-package see the file NEWS.
 
-;; $Id: tree-buffer.el,v 1.4 2003/03/27 16:13:43 berndl Exp $
-
 ;;; Code:
 
 (eval-when-compile
@@ -52,6 +59,9 @@
 (silentcomp-defun button-release-event-p)
 (silentcomp-defun button-press-event-p)
 (silentcomp-defun event-key)
+(silentcomp-defun extent-end-position)
+(silentcomp-defun event-glyph-extent)
+(silentcomp-defun event-over-glyph-p)
 (silentcomp-defun display-message)
 (silentcomp-defun clear-message)
 ;; Emacs
@@ -63,6 +73,14 @@
 (silentcomp-defun event-start)
 (silentcomp-defun posn-point)
 (silentcomp-defun event-basic-type)
+(silentcomp-defun display-graphic-p)
+
+(defconst tree-buffer-images-can-be-used
+  (and (or (fboundp 'defimage)
+           (fboundp 'make-image-specifier))
+       (if (fboundp 'display-graphic-p)
+           (display-graphic-p)
+         window-system)))
 
 
 (if tree-buffer-running-xemacs
@@ -83,6 +101,31 @@
                (ignore-errors (event-key event)))))
       (defalias 'tree-buffer-event-window 'event-window)
       (defalias 'tree-buffer-event-point 'event-point)
+      ;; stolen from dframe.el of the speedbar-library.
+      (defun tree-buffer-mouse-set-point (e)
+        "Set POINT based on event E. Handles clicking on images in XEmacs."
+        (mouse-set-point e)
+        (if (and (fboundp 'event-over-glyph-p) (event-over-glyph-p e))
+            ;; We are in XEmacs, and clicked on a picture
+            (let ((ext (event-glyph-extent e)))
+              ;; This position is back inside the extent where the
+              ;; junk we pushed into the property list lives.
+              (if (extent-end-position ext)
+                  (goto-char (1- (extent-end-position ext)))))))
+;;       (defun tree-buffer-mouse-set-point (e)
+;;         "Set POINT based on event E. Handles clicking on images in XEmacs."
+;;         (if (and (fboundp 'event-over-glyph-p) (event-over-glyph-p e))
+;;             ;; We are in XEmacs, and clicked on a picture
+;;             (let ((ext (event-glyph-extent e)))
+;;               ;; This position is back inside the extent where the
+;;               ;; junk we pushed into the property list lives.
+;;               (if (extent-end-position ext)
+;;                   (progn
+;;                     (mouse-set-point e)
+;;                     (goto-char (1- (extent-end-position ext))))
+;;                 (mouse-set-point e)))
+;;           ;; We are not in XEmacs, OR we didn't click on a picture.
+;;           (mouse-set-point e)))
       (require 'overlay)
       )
   ;; GNU Emacs
@@ -96,6 +139,7 @@
     (posn-window (event-start event)))
   (defun tree-buffer-event-point (event)
     (posn-point (event-start event)))
+  (defalias 'tree-buffer-mouse-set-point 'mouse-set-point)
   (defun tree-buffer-event-to-key (event)
     (let ((type (event-basic-type event)))
       (cond ((or (equal type 'mouse-1)
@@ -138,6 +182,7 @@ node name.")
 (defvar tree-buffer-last-incr-searchpattern nil)
 (defvar tree-buffer-incr-search nil)
 (defvar tree-buffer-hor-scroll-step nil)
+(defvar tree-buffer-expand-collapse-tokens nil)
 
 ;; tree-buffer-local data-storage with get- and set-function
 (defvar tree-buffer-data-store nil)
@@ -245,7 +290,7 @@ with the same arguments as `tree-node-expanded-fn'."
                    ;; if the expand-symbol is displayed before and mouse-button
                    ;; = 0, means RET is pressed, we do not toggle-expand but work
                    ;; as if point would not be at expand-symbol. This is for
-                   ;; conveniance.
+                   ;; convenience.
                    (not (and (= mouse-button 0)
                              tree-buffer-expand-symbol-before)))
               (progn
@@ -331,11 +376,32 @@ with the same arguments as `tree-node-expanded-fn'."
          (< (- (current-column) (window-hscroll window))
             (window-width window)))))
 
-(defun tree-buffer-scroll-hor (amount)
+(defun tree-buffer-hscroll (amount)
   (ignore-errors
     (let ((current-prefix-arg amount))
       (call-interactively 'scroll-left))))
 
+;; Stolen from dframe.el from the speedbar-library
+;; XEmacs: this can be implemented using modeline key-maps, but there
+;; is no use, as we have horizontal scrollbar (as the docstring
+;; hints.)
+(defun tree-buffer-mouse-hscroll (e)
+  "Read a mouse event E from the mode line and scroll horizontally.
+If the mouse is being clicked on the far left, or far right of the
+mode-line.  This is only useful for non-XEmacs"
+  (interactive "e")
+  (let* ((x-point (car (nth 2 (car (cdr e)))))
+	 (pixels-per-10-col (/ (* 10 (frame-pixel-width))
+			       (frame-width)))
+	 (click-col (1+ (/ (* 10 x-point) pixels-per-10-col)))
+	 )
+    (cond ((< click-col 3)
+	   (tree-buffer-hscroll (- tree-buffer-hor-scroll-step)))
+	  ((> click-col (- (window-width) 4))
+	   (tree-buffer-hscroll tree-buffer-hor-scroll-step))
+          (t (tree-buffer-nolog-message
+	      "Click on the edge of the modeline to scroll left/right")))
+    ))
 
 (defun tree-buffer-recenter (node window)
   "If NODE is not visible then first recenter the window WINDOW so NODE is
@@ -349,7 +415,7 @@ displayed without empty-lines at the end, means WINDOW is always best filled."
          (point-lines-before (count-lines (point-min) node-point))
          (point-lines-after (1- (count-lines node-point (point-max)))))
     (if (not tree-buffer-running-xemacs)
-        (ignore-errors (tree-buffer-scroll-hor -1000)))
+        (ignore-errors (tree-buffer-hscroll -1000)))
     ;; first make point best visible, means display node in the middle of the
     ;; window if possible (if there are enough lines before/after the node).
     (when (not (pos-visible-in-window-p node-point window))
@@ -462,7 +528,7 @@ returned."
             ;; make node visible if not and optimize the windows display for
             ;; the node.
             (tree-buffer-recenter node w))
-          ;; we have highlighted the node wo we return not nil.
+          ;; we have highlighted the node so we return not nil.
           t))
     (tree-buffer-remove-highlight)
     nil))
@@ -487,23 +553,66 @@ tree-node. This is only used with GNU Emacs 21!"
 text gets this face or it can be a function-symbol which is called to face the
 inserted TEXT. Such a function gets two arguments: Point where TEXT has been
 inserted and the TEXT itself"
-  (let ((p (point)))
-    (insert text)
-    (put-text-property p (+ p (length text)) 'mouse-face 'highlight)
-    (if (and help-echo (not tree-buffer-running-xemacs))
-        (put-text-property p (+ p (length text)) 'help-echo
-                           'tree-buffer-help-echo-fn))
-    (if facer
-	(if (functionp facer)
-	    (funcall facer p text)
-	  (put-text-property p (+ p (length text)) 'face facer)))))
+  (when (stringp text)
+    (let ((p (point)))
+      (insert text)
+      (put-text-property p (+ p (length text)) 'mouse-face 'highlight)
+      (if (and help-echo (not tree-buffer-running-xemacs))
+          (put-text-property p (+ p (length text)) 'help-echo
+                             'tree-buffer-help-echo-fn))
+      (if facer
+          (if (functionp facer)
+              (funcall facer p text)
+            (put-text-property p (+ p (length text)) 'face facer))))))
+
+
+(defun tree-buffer-make-image-icon-maybe (symbol-str)
+  "Add an image button based on SYMBOL-STR to SYMBOL-STR which is a string. If
+SYMBOL-STR is unknown, do nothing. If we have an image-symbol associated with
+it, use that image-symbol. Symbol-strings are associated with image-symbols
+via `tree-buffer-expand-collapse-tokens'.
+
+Always return SYMBOL-STR."
+  (if symbol-str
+      (tree-buffer-add-image-icon-maybe
+       0 (length symbol-str) symbol-str
+       (cdr (assoc symbol-str tree-buffer-expand-collapse-tokens)))))
+      
+
+(defun tree-buffer-add-image-icon-maybe (start len str image-icon)
+  "Add IMAGE-ICON to SYMBOL-STR which is a string. If IMAGE-ICON-SYMBOL is not
+nil and has a value \(which must be an image in the sense of \(X)Emacs) then
+add this image to SYMBOL-STR otherwise do nothing. Always return SYMBOL-STR."
+  (when tree-buffer-images-can-be-used
+	;; Regular images (created with `insert-image' are intangible
+	;; which (I suppose) make them more compatible with XEmacs 21.
+	;; Unfortunately, there is a giant pile o code dependent on the
+	;; underlying text.  This means if we leave it tangible, then I
+	;; don't have to change said giant piles o code.
+	(if (and image-icon (symbol-value image-icon))
+	    (if tree-buffer-running-xemacs
+		(add-text-properties (+ start len) start
+				     (list 'end-glyph (symbol-value image-icon)
+					   'rear-nonsticky (list 'display)
+					   'invisible t
+					   'detachable t)
+                                     str)
+	      (add-text-properties start (+ start len)
+				   (list 'display (symbol-value image-icon)
+					 'rear-nonsticky (list 'display))
+                                   str))))
+  str)
 
 (defun tree-buffer-add-node (node depth)
   (let* ((ww (window-width))
 	 (name (tree-node-get-name node))
 	 (width (+ (* depth tree-buffer-indent)
 		   (length name)
-		   (if (tree-node-is-expandable node) 4 0))))
+		   (if (tree-node-is-expandable node) 4 0)))
+         (expand-collapse-token
+          (if (tree-node-is-expanded node)
+              (car (nth 1 tree-buffer-expand-collapse-tokens))
+            (car (nth 0 tree-buffer-expand-collapse-tokens)))))
     ;; Truncate name if necessary
     (when (>= width ww)
       (if (eq 'beginning (tree-node-get-shorten-name node))
@@ -518,19 +627,23 @@ inserted and the TEXT itself"
     (insert (make-string (* depth tree-buffer-indent) ? ))
     (when (and tree-buffer-expand-symbol-before
 	       (tree-node-is-expandable node))
-      (tree-buffer-insert-text (if (tree-node-is-expanded node) "[-]" "[+]"))
+      (tree-buffer-insert-text 
+       (tree-buffer-make-image-icon-maybe expand-collapse-token))
       (insert " "))
     (tree-buffer-insert-text name (tree-buffer-get-node-facer node) t)
     (when (and (not tree-buffer-expand-symbol-before)
 	       (tree-node-is-expandable node))
       (insert " ")
-      (tree-buffer-insert-text (if (tree-node-is-expanded node) "[-]" "[+]")))
+      (tree-buffer-insert-text 
+       (tree-buffer-make-image-icon-maybe expand-collapse-token)))
     (insert "\n")
     (setq tree-buffer-nodes (append tree-buffer-nodes (list (cons name node))))
     (if (tree-node-is-expanded node)
 	(dolist (node (tree-node-get-children node))
 	  (tree-buffer-add-node node (1+ depth))))))
 
+
+
 (defun tree-node-count-subnodes-to-display (node)
   "Returns the number of ALL subnodes of NODE which will currently be displayed
 if NODE is expanded, means the number of all the children of NODE \(if NODE is
@@ -569,7 +682,7 @@ its current nodes. window-start and point will be preserved.
 If NODE is not nil and a valid and expanded node with at least one child then
 the display of this node is optimized so the node itself and as much as
 possible of its children \(and also recursive the children of a child if its
-aleady expanded, see `tree-node-count-subnodes-to-display') are visible in
+already expanded, see `tree-node-count-subnodes-to-display') are visible in
 current tree-buffer."
   (let* ((w (get-buffer-window (current-buffer)))
          (ws (window-start w))
@@ -627,7 +740,7 @@ level is > then LEVEL.
 
 Examples:
 - LEVEL = 0 expands only nodes which have no indentation itself.
-- LEVEL = 2 expands nodess which are either not indented or indented once or
+- LEVEL = 2 expands nodes which are either not indented or indented once or
   twice."
   (dolist (node (tree-node-get-children tree-buffer-root))
     (tree-buffer-expand-node node 0 level
@@ -664,9 +777,23 @@ see `tree-buffer-expand-nodes'."
 (defun tree-buffer-get-root ()
   tree-buffer-root)
 
-(defun tree-buffer-show-menu (event)
-  (interactive "e")
-  (mouse-set-point event)
+;; Klaus Berndl <klaus.berndl@sdm.de>: Seems that the docstring of
+;; x-popup-menu is wrong because it seems this function needs offsets related
+;; to current window not to frame!
+;; TODO: Klaus Berndl <klaus.berndl@sdm.de>: For XEmacs this does not work!
+(defun tree-buffer-show-menu-keyboard ()
+  "Activate the popup-menu of current tree-buffer via keyboard."
+  (interactive)
+  (if tree-buffer-running-xemacs
+      (tree-buffer-show-menu)
+    (let ((curr-frame-ypos (* (/ (frame-pixel-height) (frame-height))
+                              (count-lines (window-start) (point))))
+          (curr-frame-xpos (* (/ (frame-pixel-width) (frame-width))
+                              (current-column))))
+      (tree-buffer-show-menu (list (list curr-frame-xpos curr-frame-ypos)
+                                   (selected-window))))))
+
+(defun tree-buffer-show-menu (&optional event)
   (unless (not (equal (selected-frame) tree-buffer-frame))
     (when tree-buffer-menus
       (let ((node (tree-buffer-get-node-at-point)))
@@ -684,22 +811,45 @@ see `tree-buffer-expand-nodes'."
 		  (popup-menu (cons menu-title menu))
 		(let ((fn (x-popup-menu
 			   event (cons 'keymap (cons menu-title menu)))))
-		  (when fn
+                  (when fn
 		    (funcall (car fn) node)))))))))))
 
+;; (defun tree-buffer-show-menu (event)
+;;   (interactive "e")
+;;   (tree-buffer-mouse-set-point event)
+;;   (unless (not (equal (selected-frame) tree-buffer-frame))
+;;     (when tree-buffer-menus
+;;       (let ((node (tree-buffer-get-node-at-point)))
+;; 	(when node
+;; 	  (let* ((menu (cdr (assoc (tree-node-get-type node) tree-buffer-menus)))
+;;                  (menu-title-creator
+;;                   (cdr (assoc (tree-node-get-type node) tree-buffer-menu-titles)))
+;;                  (menu-title (cond ((stringp menu-title-creator)
+;;                                     menu-title-creator)
+;;                                    ((functionp menu-title-creator)
+;;                                     (funcall menu-title-creator node))
+;;                                    (t "ECB-tree-buffer-menu"))))
+;;             (when menu
+;; 	      (if tree-buffer-running-xemacs
+;; 		  (popup-menu (cons menu-title menu))
+;; 		(let ((fn (x-popup-menu
+;; 			   event (cons 'keymap (cons menu-title menu)))))
+;;                   (when fn
+;; 		    (funcall (car fn) node)))))))))))
+
 (defconst tree-buffer-incr-searchpattern-basic-prefix
-  "^[ \t]*\\(\\[[+-]\\] \\)?"
+  "^[ \t]*\\([[<][+-][]>] \\)?"
   "Prefix-pattern which ignores all not interesting basic stuff of a displayed
 token at incr. search. The following contents of a displayed token are ignored
 by this pattern:
 - beginning spaces
-- The expand/collapse-button: \[+] resp. \[-]")
+- The expand/collapse-buttons: \[+], <+> rsp. \[-], <->")
 
 (defconst tree-buffer-incr-searchpattern-node-prefix "\\([^ ]+ \\|[-+#]\\)?"
   "Prefix-pattern which ignores all not interesting stuff of a node-name at
 incr. search. The following contents of a node-name are ignored by this
 pattern:
-- types of a variable or returntypes of a method
+- types of a variable or return-types of a method
 - const specifier of variables
 - protection sign of a variable/method: +, - or #")
 
@@ -738,14 +888,15 @@ pattern:
     result))
 
 (defun tree-buffer-incremental-node-search ()
-  "Incremental search for a node in current tree-buffer. Each displayable
-key \(e.g. all keys normally bound to `self-insert-command') is appended to
-the current seach-pattern. The tree-buffer tries to jump to the current
-search-pattern. If no match is found then nothing is done. Some special keys:
+  "Incremental search for a node in current tree-buffer.
+Each display-able key \(e.g. all keys normally bound to `self-insert-command')
+is appended to the current search-pattern. The tree-buffer tries to jump to
+the current search-pattern. If no match is found then nothing is done. Some
+special keys:
 - \[backspace] and \[delete]: Delete the last character from the search-pattern.
 - \[home]: Delete the complete search-pattern
 - \[end]: Expand either to a complete node if current search-pattern is
-         already unique or expands to the greates common prefix of the nodes.
+         already unique or expands to the greatest common prefix of the nodes.
          If there are at least two nodes with the same greatest common-prefix
          than every hit of \[end] jumps to the next node with this common
          prefix.
@@ -819,6 +970,7 @@ mentioned above!"
              (aset v 0 (caar menu-items))
              (aset v 1 (list (cadar menu-items)
                              '(tree-buffer-get-node-at-point)))
+             (aset v 2 t)
              v))
        (cons (cadar menu-items)
 	     (cons (caar menu-items) t)))
@@ -863,18 +1015,18 @@ mentioned above!"
 
 (defvar tree-buffer-uncompleted-keyseq nil
   "Not nil only if there is at evaluation-time of this variable an uncompleted
-keysequence, e.g. the \"C-h\" of the keysequence \"C-h v\".")
+key sequence, e.g. the \"C-h\" of the key sequence \"C-h v\".")
 
 (defun tree-buffer-do-mouse-tracking ()
   "This function is called every time Emacs is idle for seconds defined in
 `tree-buffer-track-mouse-idle-delay'. It enables mouse-tracking but only if
-isearch is not active and if no uncompleted keysequence is open, means if this
-function is called by the idle timer during a keysequence is inserted by the
-user \(e.g. between the \"C-h\" and the \"v\" of the keysequence \"C-h v\"),
-then mouse-tracking is always not enabled, because otherwise all very slighly
+isearch is not active and if no uncompleted key sequence is open, means if this
+function is called by the idle timer during a key sequence is inserted by the
+user \(e.g. between the \"C-h\" and the \"v\" of the key sequence \"C-h v\"),
+then mouse-tracking is always not enabled, because otherwise all very slightly
 \(invisible) and unintended mouse-movements \(can occur for example only by
 the convulsion cause of hitting keys onto the keyboard!) would break the
-keysequence!"
+key sequence!"
   (setq track-mouse nil)
   (if (not (equal (tree-buffer-event-to-key last-input-event)
                   'mouse-movement))
@@ -953,6 +1105,7 @@ functionality is done with the `help-echo'-property and the function
 ;; pressed keys
 
 (defun tree-buffer-tab-pressed ()
+  "Perform the defined action after a TAB-hit."
   (interactive)
   (unless (not (equal (selected-frame) tree-buffer-frame))
     (let ((node (tree-buffer-get-node-at-point)))
@@ -960,8 +1113,7 @@ functionality is done with the `help-echo'-property and the function
 	(when (and tree-node-expanded-fn
 		   (not (tree-node-is-expanded node)))
 	  (funcall tree-node-expanded-fn node 0 nil nil (buffer-name)))
-	(when (tree-node-is-expandable node)
-	  (tree-node-toggle-expanded node))
+        (tree-node-toggle-expanded node)
 	;; Update the tree-buffer with optimized display of NODE           
 	(tree-buffer-update node)))))
 
@@ -972,35 +1124,34 @@ functionality is done with the `help-echo'-property and the function
     (tree-buffer-select 0 shift-pressed control-pressed)))
 
 (defun tree-buffer-arrow-pressed ()
+  "Perform smart arrow-key navigation/movement."
   (interactive)
   (unless (not (equal (selected-frame) tree-buffer-frame))
     (let ((node (tree-buffer-get-node-at-point))
-          (arrow-key (tree-buffer-event-to-key last-command-event))
-          node-expanded-p)
-      (if (tree-node-is-expandable node)
-          (progn
-            (setq node-expanded-p (tree-node-is-expanded node))
-            (cond ((equal arrow-key 'right)
-                   (if (not node-expanded-p)
-                       (tree-buffer-tab-pressed)))
-                  ((equal arrow-key 'left)
-                   (if node-expanded-p
-                       (tree-buffer-tab-pressed)
-                     ;; jump to next higher node
-                     (let* ((indent (tree-buffer-get-node-indent node))
-                            (new-indent (max 0 (- indent tree-buffer-indent)))
-                            (search-string
-                             (concat "^"
-                                     (buffer-substring
-                                      (tree-buffer-line-beginning-pos)
-                                      (+ (tree-buffer-line-beginning-pos)
-                                          new-indent))
-                                     "[^ \t]")))
-                       (re-search-backward search-string nil t))))))
-        (cond ((equal arrow-key 'left)
-               (forward-char -1))
-              ((equal arrow-key 'right)
-               (forward-char +1)))))))
+          (arrow-key (tree-buffer-event-to-key last-command-event)))
+      (cond ((equal arrow-key 'right)
+             (if (and (tree-node-is-expandable node)
+                      (not (tree-node-is-expanded node)))
+                 (tree-buffer-tab-pressed)
+               ;; jump to the first subnode
+               (forward-line 1)
+               (back-to-indentation)))
+            ((equal arrow-key 'left)
+             (if (tree-node-is-expanded node)
+                 (tree-buffer-tab-pressed)
+               ;; jump to next higher node
+               (let* ((indent (tree-buffer-get-node-indent node))
+                      (new-indent (max 0 (- indent tree-buffer-indent)))
+                      (search-string
+                       (concat "^"
+                               (buffer-substring
+                                (tree-buffer-line-beginning-pos)
+                                (+ (tree-buffer-line-beginning-pos)
+                                   new-indent))
+                               "[^ \t]")))
+                 (re-search-backward search-string nil t)
+                 (back-to-indentation))))))))
+
 
 ;; tree-buffer creation
 
@@ -1009,6 +1160,7 @@ functionality is done with the `help-echo'-property and the function
                                 node-data-equal-fn
                                 menus menu-titles tr-lines read-only tree-indent
                                 incr-search arrow-navigation hor-scroll
+                                expand-collapse
                                 &optional type-facer expand-symbol-before
                                 highlight-node-face general-face
                                 after-create-hook)
@@ -1020,7 +1172,7 @@ Before using the accessor-functions above the tree-buffer has to be the
 current buffer!
 
 NAME: Name of the buffer FRAME: Frame in
-which the tree-buffer is displayed and valid. All keybindings
+which the tree-buffer is displayed and valid. All key-bindings
        and interactive functions of the tree-buffer work only if called in
        FRAME otherwise nothing is done!
 IS-CLICK-VALID-FN: `tree-buffer-create' rebinds down-mouse-1, down-mouse-2,
@@ -1040,7 +1192,7 @@ IS-CLICK-VALID-FN: `tree-buffer-create' rebinds down-mouse-1, down-mouse-2,
                    is accepted. If the function returns nil then really
                    nothing is done by the tree-buffer after this click/hit!
 NODE-SELECTED-FN: Function to call if a node has been selected
-                  This function is called with the following paramters:
+                  This function is called with the following parameters:
                   - node: The selected node
                   - mouse-button \(0 = RET, 1 = mouse-1, 2 = mouse 2)
                   - shift-pressed
@@ -1049,7 +1201,7 @@ NODE-SELECTED-FN: Function to call if a node has been selected
                   For the last four arguments see the description above.
 NODE-EXPANDED-FN: Function to call if a node is expandable, point stays onto
                   the expand-symbol and node is not already expanded. This
-                  function is called with the following paramters:
+                  function is called with the following parameters:
                   - node: The selected node
                   - mouse-button \(0 = TAB, 1 = mouse-1, 2 = mouse 2)
                   - shift-pressed
@@ -1078,20 +1230,27 @@ MENUS: Nil or a list of one to three conses, each cons for a node-type \(0, 1
        or 2) Example: \(\(0 . menu-for-type-0) \(1 . menu-for-type-1)). The
        cdr of a cons must be a menu.
 MENU-TITLES: Nil or a list of one to three conses, each cons for a node-type
-             \(0, 1 or 2). See MENUES. The cdr of a cons must be either a
+             \(0, 1 or 2). See MENUS. The cdr of a cons must be either a
              string or a function which will be called with current node
              under point and must return a string which is displayed as the
              menu-title.
 TR-LINES: Should lines in this tree buffer be truncated \(not nil)
 READ-ONLY: Should the treebuffer be read-only \(not nil)
 TREE-INDENT: spaces subnodes should be indented.
-INCR-SEARCH: Should the incremental search be anabled in the tree-buffer.
+INCR-SEARCH: Should the incremental search be enabled in the tree-buffer.
              Three choices: 'prefix, 'substring, nil. See
              `tree-buffer-incremental-node-search'.
 ARROW-NAVIGATION: If not nil then smart navigation with horizontal arrow keys.
 HOR-SCROLL: Number of columns a hor. scroll in the tree-buffer should scroll.
             If not nil then M-mouse-1 and M-mouse-2 scroll left and right and
             also M-<left-arrow> and M-<right-arrow>.
+EXPAND-COLLAPSE: A 2-elem list where the first elem is a cons cell where the
+                 car is the token-string for the expand-symbol \(e.g.
+                 \"\[+]\") and the cdr is a symbol which has as value an image
+                 in the sense of \(X)Emacs. The second elem is the same for
+                 the collapse symbol. Both elems can be nil; then no expand-
+                 or collapse-symbol is drawn. If only the cdr of an elem is
+                 nil then no image-icon is used but always the token-string.
 TYPE-FACER: Nil or a list of one or two conses, each cons for a node-type \(0
             or 1). The cdr of a cons can be:
             - a symbol of a face
@@ -1110,7 +1269,7 @@ GENERAL-FACE: General face in which the whole tree-buffer should be displayed.
 AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
                    called directly after creating the tree-buffer and defining
                    it's local keymap. For example such a function can add
-                   additional keybindings for this tree-buffer local keymap."
+                   additional key-bindings for this tree-buffer local keymap."
   (let ((nop (function (lambda() (interactive))))
         (a-c-h (if (functionp after-create-hook)
                    (list after-create-hook)
@@ -1142,7 +1301,8 @@ AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
     (make-local-variable 'tree-buffer-last-incr-searchpattern)
     (make-local-variable 'tree-buffer-incr-search)
     (make-local-variable 'tree-buffer-hor-scroll-step)
-
+    (make-local-variable 'tree-buffer-expand-collapse-tokens)
+    
     ;; initialize the user-data-storage for this tree-buffer.
     (set (make-local-variable 'tree-buffer-data-store) nil)
 
@@ -1175,6 +1335,7 @@ AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
     (setq tree-buffer-last-incr-searchpattern "")
     (setq tree-buffer-incr-search incr-search)
     (setq tree-buffer-hor-scroll-step hor-scroll)
+    (setq tree-buffer-expand-collapse-tokens expand-collapse)
 
     ;; set a special syntax table for tree-buffers
     (set-syntax-table tree-buffer-syntax-table)
@@ -1225,24 +1386,23 @@ AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
       (if tree-buffer-running-xemacs '(button1) [down-mouse-1])
       (function (lambda(e)
 		  (interactive "e")
-                  (mouse-set-point e)
+                  (tree-buffer-mouse-set-point e)
                   (tree-buffer-select 1 nil nil))))
   
     (define-key tree-buffer-key-map
       (if tree-buffer-running-xemacs '(shift button1) [S-down-mouse-1])
       (function (lambda(e)
 		  (interactive "e")
-                  (mouse-set-point e)
+                  (tree-buffer-mouse-set-point e)
                   (tree-buffer-select 1 t nil))))
 
     (define-key tree-buffer-key-map
       (if tree-buffer-running-xemacs '(control button1) [C-down-mouse-1])
       (function (lambda(e)
 		  (interactive "e")
-                  (mouse-set-point e)
+                  (tree-buffer-mouse-set-point e)
                   (tree-buffer-select 1 nil t))))
 
-    
     (define-key tree-buffer-key-map [drag-mouse-1] nop)
     (define-key tree-buffer-key-map [mouse-1] nop)
     (define-key tree-buffer-key-map [double-mouse-1] nop)
@@ -1253,21 +1413,21 @@ AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
       (if tree-buffer-running-xemacs '(button2) [down-mouse-2])
       (function (lambda(e)
 		  (interactive "e")
-                  (mouse-set-point e)
+                  (tree-buffer-mouse-set-point e)
                   (tree-buffer-select 2 nil nil))))
 
     (define-key tree-buffer-key-map
       (if tree-buffer-running-xemacs '(shift button2) [S-down-mouse-2])
       (function (lambda(e)
 		  (interactive "e")
-                  (mouse-set-point e)
+                  (tree-buffer-mouse-set-point e)
                   (tree-buffer-select 2 t nil))))
 
     (define-key tree-buffer-key-map
       (if tree-buffer-running-xemacs '(control button2) [C-down-mouse-2])
       (function (lambda(e)
 		  (interactive "e")
-                  (mouse-set-point e)
+                  (tree-buffer-mouse-set-point e)
                   (tree-buffer-select 2 nil t))))
 
     (define-key tree-buffer-key-map [mouse-2] nop)
@@ -1277,7 +1437,13 @@ AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
     ;; mouse-3
     (define-key tree-buffer-key-map
       (if tree-buffer-running-xemacs '(button3) [down-mouse-3])
-      'tree-buffer-show-menu)
+      (function (lambda(e)
+		  (interactive "e")
+                  (tree-buffer-mouse-set-point e)
+                  (tree-buffer-show-menu e))))
+    (define-key tree-buffer-key-map (kbd "M-m")
+      'tree-buffer-show-menu-keyboard)
+    
     (define-key tree-buffer-key-map [mouse-3] nop)
     (define-key tree-buffer-key-map [double-mouse-3] nop)
     (define-key tree-buffer-key-map [triple-mouse-3] nop)
@@ -1289,29 +1455,35 @@ AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
         [M-down-mouse-1]
         (function (lambda(e)
                     (interactive "e")
-                    (mouse-set-point e)
-                    (tree-buffer-scroll-hor (- tree-buffer-hor-scroll-step)))))
+                    (tree-buffer-mouse-set-point e)
+                    (tree-buffer-hscroll (- tree-buffer-hor-scroll-step)))))
 
       (define-key tree-buffer-key-map
         [M-down-mouse-3]
         (function (lambda(e)
                     (interactive "e")
-                    (mouse-set-point e)
-                    (tree-buffer-scroll-hor tree-buffer-hor-scroll-step))))
+                    (tree-buffer-mouse-set-point e)
+                    (tree-buffer-hscroll tree-buffer-hor-scroll-step))))
       
       (define-key tree-buffer-key-map
         [C-M-down-mouse-1]
         (function (lambda(e)
                     (interactive "e")
-                    (mouse-set-point e)
-                    (tree-buffer-scroll-hor (- (- (window-width) 2))))))
+                    (tree-buffer-mouse-set-point e)
+                    (tree-buffer-hscroll (- (- (window-width) 2))))))
       
       (define-key tree-buffer-key-map
         [C-M-down-mouse-3]
         (function (lambda(e)
                     (interactive "e")
-                    (mouse-set-point e)
-                    (tree-buffer-scroll-hor (- (window-width) 2)))))
+                    (tree-buffer-mouse-set-point e)
+                    (tree-buffer-hscroll (- (window-width) 2)))))
+      
+      ;; This lets the user scroll as if we had a scrollbar...
+      (define-key tree-buffer-key-map
+        [mode-line mouse-1] 'tree-buffer-mouse-hscroll)
+      (define-key tree-buffer-key-map
+        [mode-line mouse-2] 'tree-buffer-mouse-hscroll)
       
       (define-key tree-buffer-key-map [M-mouse-1] nop)
       (define-key tree-buffer-key-map [M-mouse-3] nop)
@@ -1331,7 +1503,7 @@ AFTER-CREATE-HOOK: A function or a list of functions \(with no arguments)
   "Destroy the tree-buffer"
   (when buffer
     (setq tree-buffers (delq (get-buffer buffer) tree-buffers))
-    (kill-buffer buffer)))
+    (ignore-errors (kill-buffer buffer))))
 
 ;;; Tree node
 
